"""開催データに紐づくデータを結合した状態でロードする"""
from typing import List
from sqlalchemy.orm import joinedload
from jrdb_model import KaisaiData, app


def get_kaisais(start: int, end: int) -> List[KaisaiData]:
    """指定した期間の開催データを取得する

    Args:
        start (int): YYYYMMDD
        end (int): YYYYMMDD

    Returns:
        list[KaisaiData]: 開催データ一覧
    """
    with app.app_context():
        return (
            KaisaiData.query.filter(KaisaiData.ymd >= start, KaisaiData.ymd <= end)
            .options(joinedload("*"))
            .all()
        )



"""推論データを生成する"""
from typing import List, Any
from dataclasses import dataclass
from jrdb_model import (
    PredictData,
    KaisaiData,
)
from ..machine_learning.input_creator import InputCreator


@dataclass
class PredictFactory:
    """推論データ生成クラス"""

    kaisais: List[KaisaiData]
    model: Any  # Kerasのモデルインスタンス

    def create(self) -> List[PredictData]:
        """推論データを生成する

        Returns:
            List[PredictData]: 未登録の推論データ一覧
        """
        input_data = InputCreator(self.kaisais)
        preds = self.model.predict(input_data.x_data)
        w = 0
        entities = []
        for k in self.kaisais:
            for r in k.races:
                for h in r.racehorses:
                    hn = h.num - 1
                    pp1 = preds[0][w][hn]
                    pp2 = preds[1][w][hn]
                    pp3 = preds[2][w][hn]
                    rentai_rate = 1 - ((1 - pp1) * (1 - pp2))
                    fukusho_rate = 1 - ((1 - pp1) * (1 - pp2) * (1 - pp3))
                    entities.append(
                        PredictData(
                            racehorsekey=h.racehorsekey,
                            pp_icchaku=float(pp1),
                            pp_nichaku=float(pp2),
                            pp_sanchaku=float(pp3),
                            rentai_rate=rentai_rate,
                            fukusho_rate=fukusho_rate,
                            tansho_odds=1 / pp1,
                            fukusho_odds=1 / fukusho_rate,
                        )
                    )
                w += 1
        return entities



"""カテゴリ変数を取得する"""
from typing import List
import numpy as np
from jrdb_model import KaisaiData, BangumiData, RacehorseData
from .jrdbdummies import CategoryGetter


def getCategoryData(kaisais: List[KaisaiData]) -> np.ndarray:
    categories = _getCategories(kaisais)
    return _convertToMatrix(categories)


def _convertToMatrix(categories: List[np.ndarray]) -> np.ndarray:
    matrix = np.zeros([len(categories), 18, len(categories[0][0])])
    for raceNum in range(len(categories)):
        for horseNum in range(len(categories[raceNum])):
            matrix[raceNum][horseNum] = categories[raceNum][horseNum]
    return matrix


def _getCategories(kaisais: List[KaisaiData]) -> List[np.ndarray]:
    return [
        _getRaceCategories(kaisai, race) for kaisai in kaisais for race in kaisai.races
    ]


def _getRaceCategories(kaisai: KaisaiData, race: BangumiData) -> np.ndarray:
    horses = sorted(race.racehorses, key=lambda h: h.racehorsekey)
    for horse in horses:
        if horse.num == 1:
            dummies = _getCategory(kaisai, race, horse)
        else:
            dummies = np.vstack((dummies, _getCategory(kaisai, race, horse)))
    return dummies


def _getCategory(
    kaisai: KaisaiData, race: BangumiData, horse: RacehorseData
) -> np.ndarray:
    cg = CategoryGetter()
    return np.hstack(
        (
            cg.getTennatsu(kaisai.tennatsu),
            cg.getDistance(race.distance),
            cg.getBacode(horse.bacode),
            cg.getNum(horse.num),
            cg.getWaku(horse.waku),
            cg.getTorikeshi(_filterStrToInt(horse.torikeshi)),
            cg.getBanushikaicode(_filterStrToInt(horse.banushikai_code)),
            cg.getTraintype(_filterStrToInt(horse.trainanalysis.train_type)),
        )
    )


def _filterStrToInt(value):
    return 0 if isinstance(value, str) else value



"""推論に必要なデータセットを作成する"""
from typing import List
import numpy as np
from jrdb_model import KaisaiData
from . import training_tool
from . import category_data


class InputCreator:
    """KaisaiDataを元にndarrayを作成する"""

    def __init__(self, kaisais: List[KaisaiData]):
        score_data = training_tool.createScoreDataMatrix(kaisais)
        sds = training_tool.standardize(score_data)
        cd = category_data.getCategoryData(kaisais)
        self.x_data = np.concatenate((sds, cd), axis=2)



"""Yahoo競馬のクローラー結合テスト"""
from asyncio import run
from src.jra_tools.database.schedule import open_days


def test_open_days():
    """対象年、月の開催日をWebサイトに接続してテストする"""
    days = run(open_days(1, 2018))
    assert len(days) == 9
    assert days == [
        20180106,
        20180107,
        20180108,
        20180113,
        20180114,
        20180120,
        20180121,
        20180127,
        20180128,
    ]



"""パッケージから直接インポート可能なモジュール"""
from .database.bao_importer import *
from .database.db_checker import *
from .database.schedule import *
from .database.predict_factory import *
from .database.predict_race_factory import *

from .machine_learning.category_data import *
from .machine_learning.input_creator import *
from .machine_learning.jrdbdummies import *
from .machine_learning.kaisai_creator import *
from .machine_learning.label_creator import *
from .machine_learning.labelutil import *
from .machine_learning.training_tool import *



"""トレーニングに利用する関数"""
from typing import List
import numpy as np
from functools import reduce
from sklearn.preprocessing import StandardScaler
from jrdb_model import KaisaiData, BangumiData, RacehorseData


def createScoreDataMatrix(kaisais: List[KaisaiData]) -> np.ndarray:
    num_max_horse = 18
    num_race = reduce(lambda x, y: x + len(y.races), kaisais, 0)
    num_score = numberOfScoreFeatures(kaisais[0])
    baseMatrix = np.zeros([num_race, num_max_horse, num_score])
    return _setScores(baseMatrix, kaisais)


def _setScores(score_data: np.ndarray, kaisais: List[KaisaiData]) -> np.ndarray:
    w = 0
    for kaisai in kaisais:
        for race in kaisai.races:
            scores = _addKaisaiScores([], kaisai)
            score_data = _setRaceScores(score_data, race, w, scores)
            w = w + 1
    return score_data


def _setRaceScores(
    matrix: np.ndarray, race: BangumiData, raceNum: int, kaisaiScores
) -> np.ndarray:
    raceScores = kaisaiScores
    raceScores.append(race.num_of_all_horse)
    for horse in race.racehorses:
        scores = _addHorseScores(raceScores, horse)
        matrix = _setScoreData(matrix, raceNum, horse.num - 1, scores)
    return matrix


def standardize(matrix: np.ndarray) -> np.ndarray:
    sds = matrix
    for i in range(len(matrix)):
        ss = StandardScaler()
        ss.fit(matrix[i])
        sds[i] = ss.transform(matrix[i])
    return sds


def numberOfScoreFeatures(kaisai: KaisaiData) -> int:
    dummyScores = _addKaisaiScores([], kaisai)
    dummyScores.append(kaisai.races[0].num_of_all_horse)
    dummyScores = _addHorseScores(dummyScores, kaisai.races[0].racehorses[0])
    return len(dummyScores)


def _setScoreData(
    matrix: np.ndarray, raceNum: int, horseNum: int, scores: List
) -> np.ndarray:
    features = len(matrix[raceNum, horseNum])
    if features != len(scores):
        raise RuntimeError("特徴量の数と行列の次元数が一致しません")
    for s in range(features):
        matrix[raceNum, horseNum, s] = scores[s]
    return matrix


def _addKaisaiScores(scores: List, kaisai: KaisaiData) -> List:
    return scores + [
        kaisai.turf_baba_in,
        kaisai.turf_baba_center,
        kaisai.turf_baba_out,
        kaisai.turf_baba_sa,
        kaisai.turf_baba_straight_saiuchi,
        kaisai.turf_baba_straight_in,
        kaisai.turf_baba_straight_center,
        kaisai.turf_baba_straight_out,
        kaisai.turf_baba_straight_oosoto,
        kaisai.dart_baba_in,
        kaisai.dart_baba_center,
        kaisai.dart_baba_out,
        kaisai.dart_baba_sa,
        kaisai.renzoku_day,
        kaisai.turf_length,
        kaisai.precipitation,
    ]


def _addHorseScores(scores: List, horse: RacehorseData) -> List:
    horseScores = [
        horse.idm,
        horse.jockey_score,
        horse.info_score,
        horse.routin,
        horse.train_score,
        horse.trainer_score,
        horse.train_code,
        horse.trainer_hyoka_code,
        horse.jockey_rate_rentai,
        horse.gekiso_score,
        horse.kinryo,
        horse.kakutoku_money,
        horse.shukaku_money,
        horse.joken,
        horse.ten_score,
        horse.pace_score,
        horse.up_score,
        horse.position_score,
        horse.commit_weight,
        horse.commit_weight_increase,
        horse.gekiso_order,
        horse.ls_score_order,
        horse.ten_score_order,
        horse.pace_score_order,
        horse.up_score_order,
        horse.position_score_order,
        horse.expect_jokey_win_rate,
        horse.expect_jokey_rentai_rate,
        horse.taikei,
        horse.senaka,
        horse.do,
        horse.siri,
        horse.tomo,
        horse.harabukuro,
        horse.head,
        horse.neck,
        horse.breast,
        horse.shoulder,
        horse.zencho,
        horse.kocho,
        horse.maehaba,
        horse.ushirohaba,
        horse.maetsunagi,
        horse.ushirotsunagi,
        horse.tail,
        horse.furikata,
        horse.horse_start_score,
        horse.horse_latestart_rate,
        horse.mambaken_score,
        horse.runtimes_first_train,
        horse.days_after_first_train,
        horse.trainer_rank,
        horse.trainanalysis.oikiri_score,
        horse.trainanalysis.shiage_score,
        horse.trainoikiri.kaisu,
        horse.trainoikiri.train_f,
        horse.trainoikiri.ten_f,
        horse.trainoikiri.mid_f,
        horse.trainoikiri.end_f,
        horse.trainoikiri.ten_f_score,
        horse.trainoikiri.mid_f_score,
        horse.trainoikiri.end_f_score,
        horse.trainoikiri.oikiri_score,
    ]
    return scores + _filterStringToInt(horseScores)


def _filterStringToInt(scores: List) -> List[int]:
    return [0 if type(score) is str else score for score in scores]



"""Yahoo競馬のクローラーテスト"""
import pickle
from typing import Generator
import pytest
from aiohttp.client_exceptions import ClientConnectionError
from aioresponses import aioresponses
from src.jra_tools.database.schedule import open_days, annual_schedule


@pytest.fixture
def server() -> Generator[aioresponses, None, None]:
    """リクエストを受け付けるサーバーをモックする

    Yields:
        Generator[aioresponses, None, None]: _description_
    """
    with aioresponses() as m:
        with open("jra_tools/tests/yahoo_keiba.pkl", mode="rb") as f:
            m.get(
                "https://sports.yahoo.co.jp/keiba/schedule/monthly?month=12&year=2018",
                status=200,
                body=pickle.load(f),
            )
            yield m


@pytest.mark.asyncio
async def test_open_days(server: aioresponses):
    """対象の年、月に対するユニットテスト

    Args:
        server (aioresponses): モックサーバー
    """
    days = await open_days(12, 2018)

    server.assert_called_once()
    assert days == [
        20181201,
        20181202,
        20181208,
        20181209,
        20181215,
        20181216,
        20181222,
        20181223,
        20181228,
    ]


@pytest.mark.asyncio
async def test_annual_schedule(server: aioresponses):
    """
    スケジューラをテストする
    12の並列リクエストの内、一つでもコネクションができないとエラーが発生する

    Args:
        mocker (MockFixture): モッカーオブジェクト
    """

    with pytest.raises(ClientConnectionError) as e:
        await annual_schedule(2018)

    server.assert_called()
    assert (
        str(e.value)
        == "Connection refused: GET https://sports.yahoo.co.jp/keiba/schedule/monthly?month=1&year=2018"
    )
